/***********************************************************************************************************************
*
* Copyright (C) 2012 by Jim Mutch (www.jimmutch.com).
*
* This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
* License as published by the Free Software Foundation; either version 3 of the License, or (at your option) any later
* version.
*
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
* warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License along with this program.  If not, see
* <http://www.gnu.org/licenses/>.
*
***********************************************************************************************************************/

// The functions in this first section perform the basic operations needed by the various FIND_ macros.  Those macros
// are then defined in following sections.

/*--------------------------------------------------------------------------------------------------------------------*/

INLINE void _RFFindDist(float s, float d, float c, float r, int &c1, int &c2) {

    // Find all indices (valid or not) within radius R of C.

    float dd = 1.0f / d;

    c1 = (int)ceilf ((c - r - s) * dd - 0.001f);
    c2 = (int)floorf((c + r - s) * dd + 0.001f);

}

/*--------------------------------------------------------------------------------------------------------------------*/

INLINE void _RFFindNear(float s, float d, float c, int n, int &c1, int &c2) {

    // Find the N indices (valid or not) closest to C.

    float dd = 1.0f / d;

    c1 = (int)ceilf((c - s) * dd - 0.5f * (float)n - 0.001f);
    c2 = c1 + n - 1;

}

/*--------------------------------------------------------------------------------------------------------------------*/

INLINE void _RFFindNearRev(float cd, float s, float d, float c, int n, int &c1, int &c2) {

    // Inverse of _RFFindNear.  Assume cells in layer L2 are using _RFFindNear(N) to find cells in layer L1.
    // This function finds all indices (valid or not) in L2 which would find an L1 cell centered at C.

    float r  = 0.5f * (float)n * cd;
    float dd = 1.0f / d;

    c1 = (int)ceilf ((c - r - s) * dd - 0.001f);
    c2 = (int)floorf((c + r - s) * dd - 0.001f); // note minus epsilon

}

/*--------------------------------------------------------------------------------------------------------------------*/

INLINE void _RFValid(int t, int c1, int c2, int &v1, int &v2) {

    // Finds the valid range within a range of indices.

    v1 = min(max(c1,  0), t    );
    v2 = min(max(c2, -1), t - 1);

}

/*--------------------------------------------------------------------------------------------------------------------*/

INLINE void _RFFilt(int c1, int c2, int v1, int v2, int &f1, int &f2) {

    // For a filtering operation occuring near an edge, when some part of the RF is invalid, this function determines
    // the range of filter indices which should actually be used.

    f1 = v1 - c1;
    f2 = v2 - c1;

}

/*--------------------------------------------------------------------------------------------------------------------*/

INLINE void _RFFiltDistRev(int c1, int c2, int v1, int v2, int &f1, int &f2) {

    // Inverse of _RFFilt, used when we are backpropagating through a filter that used _RFFindDist.
    // Note the resulting index range is descending.

    f1 = c2 - v1;
    f2 = c2 - v2;

}

/*--------------------------------------------------------------------------------------------------------------------*/

INLINE void _RFFiltNearRev(float cd, float s, float d, float c, int n, int v1, int v2, int &f1, int &df, int &f2) {

    // Inverse of _RFFilt, used when we are backpropagating through a filter that used _RFFindNear.
    // Note the resulting index range is descending, and may decrease in steps > 1.  DF will contain the decrement.

    float cdd = 1.0f / cd;

    f1 = (int)floorf((c - (s + (float)v1 * d)) * cdd + 0.5f * (float)n);
    f2 = (int)floorf((c - (s + (float)v2 * d)) * cdd + 0.5f * (float)n);

    df = (int)floorf(d * cdd + 0.5f);

}

/*--------------------------------------------------------------------------------------------------------------------*/

INLINE bool _RFAny(int t, int c1, int c2) {

    // Returns true iff any indices in the range are valid.

    return (c1 <= c2) && (c1 < t) && (c2 >= 0);

}

/*--------------------------------------------------------------------------------------------------------------------*/

INLINE bool _RFAll(int t, int c1, int c2) {

    // Returns true iff all indices in the range are valid.

    return (c1 <= c2) && (c1 >= 0) && (c2 < t);

}

/**********************************************************************************************************************/

// FIND_type_dim_WITHIN   (Z,    R, &V1, &V2, &C1, &C2)
// FIND_type_dim_WITHIN_AT(Z, P, R, &V1, &V2, &C1, &C2)

#define _GetRFDist(Z,D1,F1,S2,D2,F2,...) \
    _GetRFDistAt(Z, S2, D2, F2, _GetCenter(D1,F1), __VA_ARGS__)

#define _GetRFDistAt(Z,S2,D2,F2,...) \
    _RFDistOrig(_GetZLayerSz(Z,S2,D2), _GetZLConst(Z,F2), _GetZLConst(Z,F2+1), __VA_ARGS__)

INLINE bool _RFDistOrig(int t, float s, float d, float c, float r, int &v1, int &v2, int &c1, int &c2) {
    _RFFindDist(s, d, c, r, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return _RFAny(t, c1, c2);
}

INLINE bool _RFDistOrig(int t, float s, float d, float c, float r, int &v1, int &v2, int &c1) {
    int c2;
    _RFFindDist(s, d, c, r, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return _RFAny(t, c1, c2);
}

INLINE bool _RFDistOrig(int t, float s, float d, float c, float r, int &v1, int &v2) {
    int c1, c2;
    _RFFindDist(s, d, c, r, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return _RFAll(t, c1, c2);
}

INLINE bool _RFDistOrig(int t, float s, float d, float c, float r, int &v1) {
    int c1, c2, v2;
    _RFFindDist(s, d, c, r, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return _RFAll(t, c1, c2);
}

INLINE int _RFDistOrig(int t, float s, float d, float c, float r) {
    int c1, c2, v1, v2;
    _RFFindDist(s, d, c, r, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return v1;
}

/*--------------------------------------------------------------------------------------------------------------------*/

// FIND_type_dim_WITHIN_FULL   (Z,    R, &C1, &C2, &V1, &V2, &F1, &F2)
// FIND_type_dim_WITHIN_FULL_AT(Z, P, R, &C1, &C2, &V1, &V2, &F1, &F2)

#define _GetRFDistFull(Z,D1,F1,S2,D2,F2,...) \
    _GetRFDistFullAt(Z, S2, D2, F2, _GetCenter(D1,F1), __VA_ARGS__)

#define _GetRFDistFullAt(Z,S2,D2,F2,...) \
    _RFDistFull(_GetZLayerSz(Z,S2,D2), _GetZLConst(Z,F2), _GetZLConst(Z,F2+1), __VA_ARGS__)

INLINE bool _RFDistFull(int t, float s, float d, float c, float r, int &c1, int &c2, int &v1, int &v2, int &f1, int &f2) {
    _RFFindDist(s, d, c, r, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    _RFFilt(c1, c2, v1, v2, f1, f2);
    return _RFAny(t, c1, c2);
}

INLINE bool _RFDistFull(int t, float s, float d, float c, float r, int &c1, int &c2, int &v1, int &v2, int &f1) {
    int f2;
    _RFFindDist(s, d, c, r, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    _RFFilt(c1, c2, v1, v2, f1, f2);
    return _RFAny(t, c1, c2);
}

INLINE bool _RFDistFull(int t, float s, float d, float c, float r, int &c1, int &c2, int &v1, int &v2) {
    _RFFindDist(s, d, c, r, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return _RFAny(t, c1, c2);
}

INLINE bool _RFDistFull(int t, float s, float d, float c, float r, int &c1, int &c2, int &v1) {
    int v2;
    _RFFindDist(s, d, c, r, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return _RFAny(t, c1, c2);
}

INLINE bool _RFDistFull(int t, float s, float d, float c, float r, int &c1, int &c2) {
    _RFFindDist(s, d, c, r, c1, c2);
    return _RFAny(t, c1, c2);
}

INLINE bool _RFDistFull(int t, float s, float d, float c, float r, int &c1) {
    int c2;
    _RFFindDist(s, d, c, r, c1, c2);
    return _RFAny(t, c1, c2);
}

INLINE int _RFDistFull(int t, float s, float d, float c, float r) {
    int c1, c2;
    _RFFindDist(s, d, c, r, c1, c2);
    return c1;
}

/*--------------------------------------------------------------------------------------------------------------------*/

// FIND_type_dim_WITHIN_VALID   (Z,    R, &V1, &V2, &F1, &F2)
// FIND_type_dim_WITHIN_VALID_AT(Z, P, R, &V1, &V2, &F1, &F2)

#define _GetRFDistValid(Z,D1,F1,S2,D2,F2,...) \
    _GetRFDistValidAt(Z, S2, D2, F2, _GetCenter(D1,F1), __VA_ARGS__)

#define _GetRFDistValidAt(Z,S2,D2,F2,...) \
    _RFDistValid(_GetZLayerSz(Z,S2,D2), _GetZLConst(Z,F2), _GetZLConst(Z,F2+1), __VA_ARGS__)

INLINE bool _RFDistValid(int t, float s, float d, float c, float r, int &v1, int &v2, int &f1, int &f2) {
    int c1, c2;
    _RFFindDist(s, d, c, r, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    _RFFilt(c1, c2, v1, v2, f1, f2);
    return _RFAll(t, c1, c2);
}

INLINE bool _RFDistValid(int t, float s, float d, float c, float r, int &v1, int &v2, int &f1) {
    int c1, c2, f2;
    _RFFindDist(s, d, c, r, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    _RFFilt(c1, c2, v1, v2, f1, f2);
    return _RFAll(t, c1, c2);
}

INLINE bool _RFDistValid(int t, float s, float d, float c, float r, int &v1, int &v2) {
    int c1, c2;
    _RFFindDist(s, d, c, r, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return _RFAll(t, c1, c2);
}

INLINE bool _RFDistValid(int t, float s, float d, float c, float r, int &v1) {
    int c1, c2, v2;
    _RFFindDist(s, d, c, r, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return _RFAll(t, c1, c2);
}

INLINE int _RFDistValid(int t, float s, float d, float c, float r) {
    int c1, c2, v1, v2;
    _RFFindDist(s, d, c, r, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return v1;
}

/**********************************************************************************************************************/

// NOTE: for the RWITHIN case, WITHIN macros can be used, unless you want filter indices.  Thus, the only RWITHIN macros
// generated are those which return filter indices.

/*--------------------------------------------------------------------------------------------------------------------*/

// FIND_type_dim_RWITHIN_FULL   (Z,    R, &C1, &C2, &V1, &V2, &F1, &F2)
// FIND_type_dim_RWITHIN_FULL_AT(Z, P, R, &C1, &C2, &V1, &V2, &F1, &F2)

#define _GetRFRevDistFull(Z,D1,F1,S2,D2,F2,...) \
    _GetRFRevDistFullAt(Z, S2, D2, F2, _GetCenter(D1,F1), __VA_ARGS__)

#define _GetRFRevDistFullAt(Z,S2,D2,F2,...) \
    _RFRevDistFull(_GetZLayerSz(Z,S2,D2), _GetZLConst(Z,F2), _GetZLConst(Z,F2+1), __VA_ARGS__)

INLINE bool _RFRevDistFull(int t, float s, float d, float c, float r, int &c1, int &c2, int &v1, int &v2, int &f1, int &f2) {
    _RFFindDist(s, d, c, r, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    _RFFiltDistRev(c1, c2, v1, v2, f1, f2);
    return _RFAny(t, c1, c2);
}

INLINE bool _RFRevDistFull(int t, float s, float d, float c, float r, int &c1, int &c2, int &v1, int &v2, int &f1) {
    int f2;
    _RFFindDist(s, d, c, r, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    _RFFiltDistRev(c1, c2, v1, v2, f1, f2);
    return _RFAny(t, c1, c2);
}

/*--------------------------------------------------------------------------------------------------------------------*/

// FIND_type_dim_RWITHIN_VALID   (Z,    R, &V1, &V2, &F1, &F2)
// FIND_type_dim_RWITHIN_VALID_AT(Z, P, R, &V1, &V2, &F1, &F2)

#define _GetRFRevDistValid(Z,D1,F1,S2,D2,F2,...) \
    _GetRFRevDistValidAt(Z, S2, D2, F2, _GetCenter(D1,F1), __VA_ARGS__)

#define _GetRFRevDistValidAt(Z,S2,D2,F2,...) \
    _RFRevDistValid(_GetZLayerSz(Z,S2,D2), _GetZLConst(Z,F2), _GetZLConst(Z,F2+1), __VA_ARGS__)

INLINE bool _RFRevDistValid(int t, float s, float d, float c, float r, int &v1, int &v2, int &f1, int &f2) {
    int c1, c2;
    _RFFindDist(s, d, c, r, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    _RFFiltDistRev(c1, c2, v1, v2, f1, f2);
    return _RFAll(t, c1, c2);
}

INLINE bool _RFRevDistValid(int t, float s, float d, float c, float r, int &v1, int &v2, int &f1) {
    int c1, c2, f2;
    _RFFindDist(s, d, c, r, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    _RFFiltDistRev(c1, c2, v1, v2, f1, f2);
    return _RFAll(t, c1, c2);
}

/**********************************************************************************************************************/

// FIND_type_dim_NEAREST   (Z,    N, &V1, &V2, &C1, &C2)
// FIND_type_dim_NEAREST_AT(Z, P, N, &V1, &V2, &C1, &C2)

#define _GetRFNear(Z,D1,F1,S2,D2,F2,...) \
    _GetRFNearAt(Z, S2, D2, F2, _GetCenter(D1,F1), __VA_ARGS__)

#define _GetRFNearAt(Z,S2,D2,F2,...) \
    _RFNearOrig(_GetZLayerSz(Z,S2,D2), _GetZLConst(Z,F2), _GetZLConst(Z,F2+1), __VA_ARGS__)

INLINE bool _RFNearOrig(int t, float s, float d, float c, int n, int &v1, int &v2, int &c1, int &c2) {
    _RFFindNear(s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return _RFAny(t, c1, c2);
}

INLINE bool _RFNearOrig(int t, float s, float d, float c, int n, int &v1, int &v2, int &c1) {
    int c2;
    _RFFindNear(s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return _RFAny(t, c1, c2);
}

INLINE bool _RFNearOrig(int t, float s, float d, float c, int n, int &v1, int &v2) {
    int c1, c2;
    _RFFindNear(s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return _RFAll(t, c1, c2);
}

INLINE bool _RFNearOrig(int t, float s, float d, float c, int n, int &v1) {
    int c1, c2, v2;
    _RFFindNear(s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return _RFAll(t, c1, c2);
}

INLINE int _RFNearOrig(int t, float s, float d, float c, int n) {
    int c1, c2, v1, v2;
    _RFFindNear(s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return v1;
}

/*--------------------------------------------------------------------------------------------------------------------*/

// FIND_type_dim_NEAREST_FULL   (Z,    N, &C1, &C2, &V1, &V2, &F1, &F2)
// FIND_type_dim_NEAREST_FULL_AT(Z, P, N, &C1, &C2, &V1, &V2, &F1, &F2)

#define _GetRFNearFull(Z,D1,F1,S2,D2,F2,...) \
    _GetRFNearFullAt(Z, S2, D2, F2, _GetCenter(D1,F1), __VA_ARGS__)

#define _GetRFNearFullAt(Z,S2,D2,F2,...) \
    _RFNearFull(_GetZLayerSz(Z,S2,D2), _GetZLConst(Z,F2), _GetZLConst(Z,F2+1), __VA_ARGS__)

INLINE bool _RFNearFull(int t, float s, float d, float c, int n, int &c1, int &c2, int &v1, int &v2, int &f1, int &f2) {
    _RFFindNear(s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    _RFFilt(c1, c2, v1, v2, f1, f2);
    return _RFAny(t, c1, c2);
}

INLINE bool _RFNearFull(int t, float s, float d, float c, int n, int &c1, int &c2, int &v1, int &v2, int &f1) {
    int f2;
    _RFFindNear(s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    _RFFilt(c1, c2, v1, v2, f1, f2);
    return _RFAny(t, c1, c2);
}

INLINE bool _RFNearFull(int t, float s, float d, float c, int n, int &c1, int &c2, int &v1, int &v2) {
    _RFFindNear(s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return _RFAny(t, c1, c2);
}

INLINE bool _RFNearFull(int t, float s, float d, float c, int n, int &c1, int &c2, int &v1) {
    int v2;
    _RFFindNear(s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return _RFAny(t, c1, c2);
}

INLINE bool _RFNearFull(int t, float s, float d, float c, int n, int &c1, int &c2) {
    _RFFindNear(s, d, c, n, c1, c2);
    return _RFAny(t, c1, c2);
}

INLINE bool _RFNearFull(int t, float s, float d, float c, int n, int &c1) {
    int c2;
    _RFFindNear(s, d, c, n, c1, c2);
    return _RFAny(t, c1, c2);
}

INLINE int _RFNearFull(int t, float s, float d, float c, int n) {
    int c1, c2;
    _RFFindNear(s, d, c, n, c1, c2);
    return c1;
}

/*--------------------------------------------------------------------------------------------------------------------*/

// FIND_type_dim_NEAREST_VALID   (Z,    N, &V1, &V2, &F1, &F2)
// FIND_type_dim_NEAREST_VALID_AT(Z, P, N, &V1, &V2, &F1, &F2)

#define _GetRFNearValid(Z,D1,F1,S2,D2,F2,...) \
    _GetRFNearValidAt(Z, S2, D2, F2, _GetCenter(D1,F1), __VA_ARGS__)

#define _GetRFNearValidAt(Z,S2,D2,F2,...) \
    _RFNearValid(_GetZLayerSz(Z,S2,D2), _GetZLConst(Z,F2), _GetZLConst(Z,F2+1), __VA_ARGS__)

INLINE bool _RFNearValid(int t, float s, float d, float c, int n, int &v1, int &v2, int &f1, int &f2) {
    int c1, c2;
    _RFFindNear(s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    _RFFilt(c1, c2, v1, v2, f1, f2);
    return _RFAll(t, c1, c2);
}

INLINE bool _RFNearValid(int t, float s, float d, float c, int n, int &v1, int &v2, int &f1) {
    int c1, c2, f2;
    _RFFindNear(s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    _RFFilt(c1, c2, v1, v2, f1, f2);
    return _RFAll(t, c1, c2);
}

INLINE bool _RFNearValid(int t, float s, float d, float c, int n, int &v1, int &v2) {
    int c1, c2;
    _RFFindNear(s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return _RFAll(t, c1, c2);
}

INLINE bool _RFNearValid(int t, float s, float d, float c, int n, int &v1) {
    int c1, c2, v2;
    _RFFindNear(s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return _RFAll(t, c1, c2);
}

INLINE int _RFNearValid(int t, float s, float d, float c, int n) {
    int c1, c2, v1, v2;
    _RFFindNear(s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return v1;
}

/**********************************************************************************************************************/

// FIND_type_dim_RNEAREST   (Z,    N, &V1, &V2, &C1, &C2)
// FIND_type_dim_RNEAREST_AT(Z, P, N, &V1, &V2, &C1, &C2)

#define _GetRFRevNear(Z,D1,F1,S2,D2,F2,...) \
    _GetRFRevNearAt(Z, F1, S2, D2, F2, _GetCenter(D1,F1), __VA_ARGS__)

#define _GetRFRevNearAt(Z,F1,S2,D2,F2,...) \
    _RFRevNearOrig(_GetLConst(F1+1), _GetZLayerSz(Z,S2,D2), _GetZLConst(Z,F2), _GetZLConst(Z,F2+1), __VA_ARGS__)

INLINE bool _RFRevNearOrig(float cd, int t, float s, float d, float c, int n, int &v1, int &v2, int &c1, int &c2) {
    _RFFindNearRev(cd, s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return _RFAny(t, c1, c2);
}

INLINE bool _RFRevNearOrig(float cd, int t, float s, float d, float c, int n, int &v1, int &v2, int &c1) {
    int c2;
    _RFFindNearRev(cd, s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return _RFAny(t, c1, c2);
}

INLINE bool _RFRevNearOrig(float cd, int t, float s, float d, float c, int n, int &v1, int &v2) {
    int c1, c2;
    _RFFindNearRev(cd, s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return _RFAll(t, c1, c2);
}

INLINE bool _RFRevNearOrig(float cd, int t, float s, float d, float c, int n, int &v1) {
    int c1, c2, v2;
    _RFFindNearRev(cd, s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return _RFAll(t, c1, c2);
}

INLINE int _RFRevNearOrig(float cd, int t, float s, float d, float c, int n) {
    int c1, c2, v1, v2;
    _RFFindNearRev(cd, s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return v1;
}

/*--------------------------------------------------------------------------------------------------------------------*/

// FIND_type_dim_RNEAREST_FULL   (Z,    N, &C1, &C2, &V1, &V2, &F1, &DF, &F2)
// FIND_type_dim_RNEAREST_FULL_AT(Z, P, N, &C1, &C2, &V1, &V2, &F1, &DF, &F2)

#define _GetRFRevNearFull(Z,D1,F1,S2,D2,F2,...) \
    _GetRFRevNearFullAt(Z, F1, S2, D2, F2, _GetCenter(D1,F1), __VA_ARGS__)

#define _GetRFRevNearFullAt(Z,F1,S2,D2,F2,...) \
    _RFRevNearFull(_GetLConst(F1+1), _GetZLayerSz(Z,S2,D2), _GetZLConst(Z,F2), _GetZLConst(Z,F2+1), __VA_ARGS__)

INLINE bool _RFRevNearFull(float cd, int t, float s, float d, float c, int n, int &c1, int &c2, int &v1, int &v2, int &f1, int &df, int &f2) {
    _RFFindNearRev(cd, s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    _RFFiltNearRev(cd, s, d, c, n, v1, v2, f1, df, f2);
    return _RFAny(t, c1, c2);
}

INLINE bool _RFRevNearFull(float cd, int t, float s, float d, float c, int n, int &c1, int &c2, int &v1, int &v2, int &f1, int &df) {
    int f2;
    _RFFindNearRev(cd, s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    _RFFiltNearRev(cd, s, d, c, n, v1, v2, f1, df, f2);
    return _RFAny(t, c1, c2);
}

INLINE bool _RFRevNearFull(float cd, int t, float s, float d, float c, int n, int &c1, int &c2, int &v1, int &v2, int &f1) {
    int df, f2;
    _RFFindNearRev(cd, s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    _RFFiltNearRev(cd, s, d, c, n, v1, v2, f1, df, f2);
    return _RFAny(t, c1, c2);
}

INLINE bool _RFRevNearFull(float cd, int t, float s, float d, float c, int n, int &c1, int &c2, int &v1, int &v2) {
    _RFFindNearRev(cd, s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return _RFAny(t, c1, c2);
}

INLINE bool _RFRevNearFull(float cd, int t, float s, float d, float c, int n, int &c1, int &c2, int &v1) {
    int v2;
    _RFFindNearRev(cd, s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return _RFAny(t, c1, c2);
}

INLINE bool _RFRevNearFull(float cd, int t, float s, float d, float c, int n, int &c1, int &c2) {
    _RFFindNearRev(cd, s, d, c, n, c1, c2);
    return _RFAny(t, c1, c2);
}

INLINE bool _RFRevNearFull(float cd, int t, float s, float d, float c, int n, int &c1) {
    int c2;
    _RFFindNearRev(cd, s, d, c, n, c1, c2);
    return _RFAny(t, c1, c2);
}

INLINE int _RFRevNearFull(float cd, int t, float s, float d, float c, int n) {
    int c1, c2;
    _RFFindNearRev(cd, s, d, c, n, c1, c2);
    return c1;
}

/*--------------------------------------------------------------------------------------------------------------------*/

// FIND_type_dim_RNEAREST_VALID   (Z,    N, &V1, &V2, &F1, &DF, &F2)
// FIND_type_dim_RNEAREST_VALID_AT(Z, P, N, &V1, &V2, &F1, &DF, &F2)

#define _GetRFRevNearValid(Z,D1,F1,S2,D2,F2,...) \
    _GetRFRevNearValidAt(Z, F1, S2, D2, F2, _GetCenter(D1,F1), __VA_ARGS__)

#define _GetRFRevNearValidAt(Z,F1,S2,D2,F2,...) \
    _RFRevNearValid(_GetLConst(F1+1), _GetZLayerSz(Z,S2,D2), _GetZLConst(Z,F2), _GetZLConst(Z,F2+1), __VA_ARGS__)

INLINE bool _RFRevNearValid(float cd, int t, float s, float d, float c, int n, int &v1, int &v2, int &f1, int &df, int &f2) {
    int c1, c2;
    _RFFindNearRev(cd, s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    _RFFiltNearRev(cd, s, d, c, n, v1, v2, f1, df, f2);
    return _RFAll(t, c1, c2);
}

INLINE bool _RFRevNearValid(float cd, int t, float s, float d, float c, int n, int &v1, int &v2, int &f1, int &df) {
    int c1, c2, f2;
    _RFFindNearRev(cd, s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    _RFFiltNearRev(cd, s, d, c, n, v1, v2, f1, df, f2);
    return _RFAll(t, c1, c2);
}

INLINE bool _RFRevNearValid(float cd, int t, float s, float d, float c, int n, int &v1, int &v2, int &f1) {
    int c1, c2, df, f2;
    _RFFindNearRev(cd, s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    _RFFiltNearRev(cd, s, d, c, n, v1, v2, f1, df, f2);
    return _RFAll(t, c1, c2);
}

INLINE bool _RFRevNearValid(float cd, int t, float s, float d, float c, int n, int &v1, int &v2) {
    int c1, c2;
    _RFFindNearRev(cd, s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return _RFAll(t, c1, c2);
}

INLINE bool _RFRevNearValid(float cd, int t, float s, float d, float c, int n, int &v1) {
    int c1, c2, v2;
    _RFFindNearRev(cd, s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return _RFAll(t, c1, c2);
}

INLINE int _RFRevNearValid(float cd, int t, float s, float d, float c, int n) {
    int c1, c2, v1, v2;
    _RFFindNearRev(cd, s, d, c, n, c1, c2);
    _RFValid(t, c1, c2, v1, v2);
    return v1;
}
